import asyncio
import contextlib
import functools
import typing

_runner = asyncio.Runner()


def as_sync[**P, T](func: typing.Callable[P, typing.Awaitable[T]]) -> typing.Callable[P, T]:
    if func is None:
        raise Exception("Deceorator does not support arguments, remove parentheses")

    @functools.wraps(func)
    def _inner(*args, **kwargs):
        return _runner.run(func(*args, **kwargs))

    return functools.update_wrapper(_inner, func)


@contextlib.contextmanager
def as_sync_context_manager[T](async_context_manager: contextlib.AbstractAsyncContextManager[T]) -> typing.Iterator[T]:
    exit_stack = contextlib.AsyncExitStack()
    try:
        yield _runner.run(exit_stack.enter_async_context(async_context_manager))
    finally:
        _runner.run(exit_stack.aclose())



import re


def match_with_wildcard(pattern: str, string: str) -> bool:
    r"""
    Matches the given string against an start and end anchored pattern containing wildcard `*` characters.

    All the other characters are interpreted literaly (does not match regexes!).
    Example:
        "abc*" matches "abcANYTHING", but not "ANYTHINGabc" nor "abc*""
        "abc[0-9]*" matches "abc[0-9]ANYTHING", not "abc1234"
    """

    pattern = re.escape(pattern)
    pattern = pattern.replace(r"\*", ".*")
    pattern = f"^{pattern}$"

    return re.match(pattern, string) is not None



import asyncio
import contextlib
import typing

from loguru import logger


class BroadcastManager[T]:
    """Broadcast manager that broadcasts an item to its open queues

    The manager supports opening read queues that receive items broadcast by :py:meth:`broadcast` or :py:meth:`broadcast_nowait`. A read queue can
    be obtained by :
        * calling :py:meth:`output` which returns an asynchronous context manager yielding a queue on enter (queue is closed on exit),
        * calling :py:meth:`open_queue` (note that :py:meth:`close_queue` must be called to close the queue).
    """

    _queues: set[asyncio.Queue[T]]

    def __init__(self):
        self._queues = set()

    def __contains__(self, item):
        return item in self._queues

    def _subscribe(self):
        logger.debug("Subscribing")
        queue = asyncio.Queue()
        self._queues.add(queue)

        return queue

    def _unsubscribe(self, queue: asyncio.Queue[T]):
        logger.debug("Unsubscribing")
        self._queues.remove(queue)

    @contextlib.contextmanager
    def output(self) -> typing.Iterator[asyncio.Queue[T]]:
        """Returns an asynchronous context manager that opens a new queue on enter and closes it on exit"""
        logger.debug("Opening output")
        queue = self._subscribe()
        try:
            yield queue
        finally:
            logger.debug("Closing output")
            self._unsubscribe(queue)

    async def broadcast(self, item: T):
        """Broadcasts the item to all open queues. Blocking variant."""
        logger.debug("Broadcasting", item=item)

        for queue in self._queues:
            await queue.put(item)

    def broadcast_nowait(self, item: T):
        """Broadcasts the item to all open queues. Non-blocking variant."""
        logger.debug("Broadcasting", item=item)

        for queue in self._queues:
            queue.put_nowait(item)



import asyncio


class RLock:
    def __init__(self) -> None:
        self._lock = asyncio.Lock()
        self._lock_owner: asyncio.Task | None = None
        self._counter = 0

    async def __aenter__(self) -> None:
        if not self._counter or self._lock_owner != asyncio.current_task():
            await self._lock.acquire()
            self._lock_owner = asyncio.current_task()

        self._counter += 1

    async def __aexit__(self, *args) -> bool:
        if self._counter == 1:
            self._lock.release()
            self._lock_owner = None

        self._counter -= 1

        return False

    def locked(self) -> bool:
        return self._lock.locked()

    def owned(self) -> bool:
        return self._lock.locked() and self._lock_owner == asyncio.current_task()



import logging
import sys
import typing

from loguru import logger

_debug_logger_format = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
    "<level>{level: <8}</level> | "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | "
    "- <level>{message} {extra}</level>"
)


def setup_logging(level: typing.Literal["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"]):
    logging.basicConfig(handlers=[_InterceptHandler()], level=0)

    logger.remove(0)
    logger.add(sys.stderr, level=level, format=_debug_logger_format, backtrace=True, diagnose=True)


class _InterceptHandler(logging.Handler):
    def emit(self, record):
        # Get corresponding Loguru level if it exists.
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())



import asyncio
import typing
from builtins import RuntimeError


async def async_iterator_never() -> typing.AsyncGenerator[typing.Never]:
    yield await asyncio.Future()


class CancellableTaskGroup(asyncio.TaskGroup):
    class _TaskGroupCancelledError(Exception): ...

    def __init__(self) -> None:
        super().__init__()
        self._suppress_cancellation_error = False

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        try:
            return await super().__aexit__(exc_type, exc_value, traceback)
        except* self._TaskGroupCancelledError:
            if not self._suppress_cancellation_error:
                raise asyncio.CancelledError() from None

    def cancel_quietly(self) -> None:
        self._suppress_cancellation_error = True
        self.cancel()

    def cancel(self) -> None:
        async def _canceller() -> None:
            raise self._TaskGroupCancelledError()

        try:
            self.create_task(_canceller())
        except RuntimeError:
            pass  # tg is already shutting down, we can ignore this



from __future__ import annotations

import functools
import operator
from typing import Any, Protocol


class _Serializable(Protocol):
    def __init__(self, val) -> None: ...
    def serialize(self) -> bytes: ...


class _Deserializable(Protocol):
    data_width: int

    @classmethod
    def deserialize(cls, data: bytes) -> _Deserializable: ...


def asciiz2str(asciiz: str | bytes) -> str:
    """Decodes ASCIIZ (C-string) to python string

    Function returns head of tring till first ``'\0'`` occourance. If ``'\0'`` not found,
    then entire string is returned. If input text is ``bytes``, then in is converted
    as ``ASCII`` to string.

    :param txt:    ASCIIZ string to be converted to python string

    :return:       Python string extracted from ASCIIZ
    """
    if isinstance(asciiz, str):
        asciiz = asciiz.encode("ascii")

    asciiz = asciiz.split(b"\x00", 1)[0]
    return asciiz.decode()


def str2asciiz(txt: str | bytes, size: int | None = None) -> bytes:
    """Encodes python string to ASCIIZ (C-string)

    Function returns python string as bytes with ``ASCII`` encoding terminated by ``b'\0'``.
    If `size` is defined, then bytes of this size will be returned containing ``ASCII``
    encoded bytes where rest of bytes is filled in by ``b'\0'`` characters. When there is not
    enough space to store encoded ``ASCIIZ``, then exception :py:class:`ProtocolSizeError`
    is raised.

    :param txt:    Python string to be converted as ``ASCIIZ``
    :param size:   Final size of buffer to store ``ASCIIZ``. Raises :py:class:`ProtocolSizeError`
                   when buffer is too small to pass whole string.

    :return:       Bytes encoded as ``ASCIIZ``
    """
    if isinstance(txt, str):
        txt = txt.encode("ascii")

    length = len(txt)

    if size is None:
        size = length + 1

    if length < size:
        txt += b"\0" * (size - length)
    elif length > size:
        raise ValueError(f"Provided string is larger than the limit: string {length}, limit: {size}")

    return txt


def deserialize_array[T: _Deserializable](member_type: type[str | T], data: bytes) -> str | list[T]:
    """Deserialize ``bytes`` to array of instances of requested type

    Takes bytes and creates array of items based on items sizes. If requested type is based on ``str``,
    then ``ASCIIZ is expected in bytes``.

    :param member_type:    Expected items type (see :py:class:`protocol_impl.Types`)
    :param data:           Bytes to be deserialized
    """
    if issubclass(member_type, str):
        return asciiz2str(data)
    elif hasattr(member_type, "py_pack") and member_type.py_pack == "B":  # type: ignore[attr-defined]
        # Array of bytes is always converted to bytes -> keep it as it is
        return data  # type: ignore[return-value]
    else:
        # The rest is converted to list of type based items
        return [
            member_type.deserialize(data[r : r + member_type.data_width])  # type: ignore[misc]
            for r in range(0, len(data), member_type.data_width)
        ]


def serialize_array[T: _Serializable](
    array: str | bytes | list[Any],
    member_type: type[str | bytes | T],
    size: int | None = None,
) -> bytes:
    """Serialize iterable of items to ``bytes``

    Takes array of items and serialize them into bytes according to size of elements.
    If an input array is ``str``, then ``ASCIIZ`` is created.

    :param array:    Iterable to be serialized
    :param size:     Size of final bytes buffer. Raises :py:meth:`ProtocolSizeError`
                     when size does not match to serialized data.
    """
    if issubclass(member_type, bytes | str):
        if isinstance(array, str | bytes):
            # Convert str to ASCIIZ
            data = str2asciiz(array, size)
        else:
            raise ValueError("Only str and bytes types can be serialized to str and bytes")
    else:
        # The rest is converted from list of type based items
        data = functools.reduce(operator.add, [member_type(x).serialize() for x in array])

    if size is not None and len(data) > size:
        raise ValueError(f"Expected array size does not match the real size, expected: {size}, real: {len(data)}")

    return data



class OzobotError(Exception):
    """Base Ozobot library error"""

    @property
    def context(self) -> dict[str, str]:
        return self._context

    def __init__(self, message: str, *, context: dict[str, str] | None = None) -> None:
        super().__init__(message)
        self._context = context or {}

    def add_context(self, key: str, value: str) -> None:
        self._context[key] = value

    def __str__(self) -> str:
        context_lines = [f"{k}={v}" for k, v in self._context.items()]
        return "\n".join([self.args[0], *context_lines])

    def __repr__(self) -> str:
        message_str = f'"{self.args[0]}"'
        if self._context:
            context_items = [f'"{k}": "{v}"' for k, v in self._context.items()]
            context_str = f"context={{{', '.join(context_items)}}}"
            args = [message_str, context_str]
        else:
            args = [message_str]
        name = self.__class__.__name__
        return f"{name}({', '.join(args)})"


class ActorError(Exception):
    """Base exception for actor errors."""


class ActorNotFoundError(ActorError):
    def __init__(self, actor: str):
        super().__init__(f"Actor not found: {actor}")


class SuitableActorNotFoundError(ActorError):
    def __init__(self, description: str):
        super().__init__(f"No suitable actor found: {description}")


class ActorAlreadyExistsError(ActorError):
    def __init__(self, actor: str):
        super().__init__(f"Actor already exists: {actor}")


class CorruptedStateError(ActorError):
    def __init__(self):
        super().__init__("Corrupted state: actor stack mismatch")



import struct


class FloatOutOfRangeError(ValueError):
    def __init__(self, val: float | int):
        super().__init__(f"Float value out of supported range (-128, 128): {val}")


def float_to_s8_24(val: float) -> int:
    if -128 < val < 128:
        return int(val * float(1 << 24))
    else:
        raise FloatOutOfRangeError(val)


def s8_24_to_float(val: int) -> float:
    return val / float(1 << 24)


def s8_24_serialize(val: float) -> bytes:
    return struct.pack("<i", float_to_s8_24(val))


def s8_24_deserialize(val: bytes) -> float:
    return float(s8_24_to_float(*struct.unpack("<i", val)))
